package cz.cuni.amis.pogamut.ut2004.examples;

import cz.cuni.amis.introspection.java.JProp;
import java.util.Collection;
import java.util.Random;
import java.util.logging.Level;

import cz.cuni.amis.pogamut.base.agent.navigation.PathEventType;
import cz.cuni.amis.pogamut.base.agent.navigation.PathExecutorListener;
import cz.cuni.amis.pogamut.base.agent.navigation.PathHandle;
import cz.cuni.amis.pogamut.base.agent.navigation.PathPlanner;
import cz.cuni.amis.pogamut.base.communication.worldview.event.IWorldEventListener;
import cz.cuni.amis.pogamut.base.utils.guice.AgentScoped;
import cz.cuni.amis.pogamut.base.utils.math.DistanceUtils;
import cz.cuni.amis.pogamut.base3d.worldview.object.ILocated;
import cz.cuni.amis.pogamut.base3d.worldview.object.Location;
import cz.cuni.amis.pogamut.ut2004.agent.navigation.UTAstar;
import cz.cuni.amis.pogamut.ut2004.agent.navigation.UTPathExecutor;
import cz.cuni.amis.pogamut.ut2004.bot.impl.UT2004Bot;
import cz.cuni.amis.pogamut.ut2004.bot.impl.UT2004BotLogicController;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbcommands.Initialize;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbcommands.Jump;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbcommands.Move;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbcommands.SendMessage;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbcommands.Stop;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.BotKilled;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.ConfigChange;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.GameInfo;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.GlobalChat;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.InitedMessage;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.NavPoint;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.Player;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.Self;
import cz.cuni.amis.pogamut.ut2004.utils.SingleUT2004BotRunner;
import cz.cuni.amis.utils.exception.PogamutException;

/**
 * Example of Simple Pogamut bot, that randomly walks around the map. Bot is uncapable 
 * of handling movers so far.
 *
 * @author Rudolf Kadlec aka ik
 */
@AgentScoped
public class NavigationBot extends UT2004BotLogicController {

    Location startLocation = new Location(4600, -1400, -3446);
    /**
     * Executor is used for following a path in the environment.
     */
    UTPathExecutor pathExecutor = null;
    /**
     * Planner used to compute the path.
     */
    PathPlanner<ILocated, ILocated> pathPlanner = null;
    /**
     * Random number generator used for picking random navpoint.
     */
    Random rand = new Random();

    @JProp
    boolean bFollowingPlayer = false;
    
    @JProp
    boolean bRunningAround = false;

    /**
     * Listener to text messages received from the environment. Called each time the bot
     * receives a message. To make this listener work it is needed to register him
     * in bots worldView!     
     */
    IWorldEventListener<GlobalChat> myTextListener = new IWorldEventListener<GlobalChat>() {

        public void notify(GlobalChat event) {

            if (event.getText().toLowerCase().contains("run")) {
                //start running around the map
                if (!bRunningAround) {
                    goToRandomNavPoint();
                    bFollowingPlayer = false;
                }
            } else if (event.getText().toLowerCase().contains("follow")) {
                // if running around the map, then disable it
                if (bRunningAround) {
                    pathExecutor.stop();
                    getAct().act(new Stop());
                    bRunningAround = false;
                }
                // set up we are ready to follow a player
                bFollowingPlayer = true;
            } else if (event.getText().toLowerCase().contains("stop")) {
                //stop everything
                pathExecutor.stop();
                bRunningAround = false;
                bFollowingPlayer = false;
                getAct().act(new Stop());
            } else if (event.getText().toLowerCase().contains("hello")) {
                //greep politely
                getAct().act(new SendMessage().setText("Hi " + event.getName() + "!").setFadeOut(10));
            } else if (event.getText().toLowerCase().contains("jump")) {
                //jump
                getAct().act(new Jump());
            }
        }
    };

    /**
     * Here we can modify initializing command for our bot.
     *
     * @return
     */
    @Override
    public Initialize getInitializeCommand() {
        Initialize myInit = new Initialize();
        myInit.setLocation(startLocation);
        return myInit;
    }

    /**
     * The bot is initilized in the environment - a physical representation of the
     * bot is present in the game.
     *
     * @param config information about configuration
     * @param init information about configuration
     */
    @Override
    public void botInitialized(GameInfo gameInfo, ConfigChange config, InitedMessage init) {
        // create new path planner
        pathPlanner = new UTAstar(bot);
        //TIP: you can try FloydWarshallPathPlanner instead of UTAstar
        //pathPlanner = new FloydWarshallPathPlanner(bot);

        // create new path executor
        pathExecutor = new UTPathExecutor(bot);
        //pathExecutor.addStuckDetector(new StupidStuckDetector(getWorldView(), 3.0));

        // specify what will happen when the bot reaches its destination or gets stuck
        pathExecutor.addPathListener(new PathExecutorListener() {

            public void onEvent(PathEventType eventType) {
                switch (eventType) {
                    case TARGET_REACHED:
                        // most of the time the execution will go this way
                        goToRandomNavPoint();

                        break;
                    case BOT_STUCKED:
                        // sometimes the bot gets stucked then we have to return him
                        // to some safe location, try more complex maps to get into
                        // this branche
                        NavPoint target = DistanceUtils.getSecondNearest(
                                getWorldView().getAll(NavPoint.class).values(),
                                bot);
                        pathExecutor.followPath(
                                pathPlanner.computePath(target));
                        break;
                }
            }
        });
        //register listener
        getWorldView().addEventListener(GlobalChat.class, myTextListener);
    }

    @Override
    public void prepareBot(UT2004Bot bot) {
        super.prepareBot(bot);
        bot.getLogger().setLevel(Level.SEVERE);
    }

    @Override
    public void botSpawned(GameInfo gameInfo, ConfigChange config,
            InitedMessage init, Self self) {
        super.botSpawned(gameInfo, config, init, self);
        // start the movement
        //goToRandomNavPoint();
    }

    /**
     * Main method that controls the bot - makes decisions what to do next.
     * It is called iteratively by Pogamut engine every time a synchronous batch
     * from the environment is received. This is usually 4 times per second - it
     * is affected by visionTime variable, that can be adjusted in GameBots ini file in
     * UT2004/System folder.
     *
     * @throws cz.cuni.amis.pogamut.base.exceptions.PogamutException
     */
    @Override
    public void logic() throws PogamutException {

        if (bFollowingPlayer) {
            Self self = getWorldView().getSingle(Self.class);
            Player player = null;
            for (Player pl : getWorldView().getAllVisible(Player.class).values()) {
                if (player == null) {
                    player = pl;
                } else if (pl.getLocation().getDistance(self.getLocation()) < player.getLocation().getDistance(self.getLocation())) {
                    player = pl;
                }
            }
            if (player != null) {
                getAct().act(new Move().setFirstLocation(player.getLocation()));
            }
        }
    }

    /**
     * Called each time our bot die. Good for reseting all bot state dependent variables.
     *
     * @param event
     */
    @Override
    public void botKilled(BotKilled event) {
        pathExecutor.stop();
    }

    /**
     * Picks random navpoint and goes to it.
     */
    protected void goToRandomNavPoint() {
        NavPoint navPoint = pickNewRandomNavTarget();
        // find path to the random navpoint, path is computed asynchronously
        // so the handle will hold the result onlt after some time
        PathHandle<ILocated> pathHandle = pathPlanner.computePath(navPoint);
        // make the path executor follow the path, executor listens for the
        // asynchronous result of path planning
        pathExecutor.followPath(pathHandle);
    }

    /**
     * Randomly picks some navigation point to head to.
     * @return randomly choosed navpoint
     */
    private NavPoint pickNewRandomNavTarget() {
        getLog().severe("Picking new target navpoint.");

        // 1. get all known navigation points
        Collection<NavPoint> navPoints = getWorldView().getAll(NavPoint.class).values();

        // 2. compute index of the target nav point
        int navPointIndex = rand.nextInt(navPoints.size());

        // 3. find the corresponding nav point
        int i = 0;
        for (NavPoint nav : navPoints) {
            if (i == navPointIndex) {
                return nav;
            }
            i++;
        }

        // 4. deal with unexpected behavior
        throw new RuntimeException(
                "No navpoint chosen. There are no navpoints in the list of known navpoints");
    }

    public static void main(String args[]) throws PogamutException {
        new SingleUT2004BotRunner<UT2004Bot>(NavigationBot.class, "NavigationBot").startAgent();
    }
}
