package cz.cuni.amis.pogamut.ut2004.examples;


import java.util.logging.Level;

import javax.vecmath.Vector3d;

import cz.cuni.amis.introspection.java.JProp;
import cz.cuni.amis.pogamut.base.utils.guice.AgentScoped;
import cz.cuni.amis.pogamut.base3d.worldview.object.Location;
import cz.cuni.amis.pogamut.ut2004.agent.module.sensomotoric.Raycasting;
import cz.cuni.amis.pogamut.ut2004.bot.command.AdvancedLocomotion;
import cz.cuni.amis.pogamut.ut2004.bot.impl.UT2004Bot;
import cz.cuni.amis.pogamut.ut2004.bot.impl.UT2004BotLogicController;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbcommands.Configuration;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbcommands.Initialize;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbcommands.RemoveRay;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.AutoTraceRay;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.BotKilled;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.ConfigChange;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.GameInfo;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.InitedMessage;
import cz.cuni.amis.pogamut.ut2004.utils.SingleUT2004BotRunner;
import cz.cuni.amis.pogamut.ut2004.utils.UnrealUtils;
import cz.cuni.amis.utils.exception.PogamutException;
import cz.cuni.amis.utils.flag.FlagListener;

/**
 * Example of Simple Pogamut bot, that randomly walks around the map. Bot is uncapable 
 * of handling movers so far.
 * @author Ondrej Burkert
 * @author Rudolf Kadlec aka ik
 */
@AgentScoped
public class RaycastingBot extends UT2004BotLogicController {

    Location startLocation = new Location(4800,-1500,-3446);
    AdvancedLocomotion loc = null;
    Raycasting raycasting = null;
    // Constants for rays' ids. It is allways better to store such values
    // in constants instead of using directly strings on multiple places of your
    // source code
    protected static final String FRONT = "frontRay";
    protected static final String LEFT45 = "left45Ray";
    protected static final String LEFT90 = "left90Ray";
    protected static final String RIGHT45 = "right45Ray";
    protected static final String RIGHT90 = "right90Ray";
    AutoTraceRay left, front, right;
    /**
     * Flag indicating that the bot has been just executed.
     */
    boolean first = true;
    boolean raysInitialized = false;
    /**
     * Whether the left sensor signalize the collision. (Computed in the doLogic())
     */
    @JProp
    boolean sensorLeft = false;
    /**
     * Whether the right sensor signalize the collision. (Computed in the doLogic())
     */
    @JProp
    boolean sensorRight = false;
    /**
     * Whether the front sensor signalize the collision. (Computed in the doLogic())
     */
    @JProp
    boolean sensorFront = false;
    /**
     * Whether the bot is moving. (Computed in the doLogic())
     */
    @JProp
    boolean moving = false;
    /**
     * Whether any of the sensor signalize the collision. (Computed in the doLogic())
     */
    @JProp
    boolean sensor = false;
    /**
     * How much time should we wait for the rotation to finish (milliseconds).
     */
    @JProp
    int turnSleep = 250;
    /**
     * How fast should we move? Iterval <0, 1>.
     */
    float moveSpeed = 0.6f;
    /**
     * Small rotation (degrees).
     */
    @JProp
    int smallTurn = 30;
    /**
     * Big rotation (degrees).
     */
    @JProp
    int bigTurn = 90;
    
    @Override
    public void prepareBot(UT2004Bot bot) {
    	super.prepareBot(bot);
    	
        bot.getLogger().setLevel(Level.INFO);
        // advanced locomotion will be used to move the bot, it is a faccade
        // for sending various commands
        loc = new AdvancedLocomotion(bot, bot.getLogger().getCategory("Locomotion"));

        // Raycasting object is utility object for initializing multiple rays at once
        raycasting = new Raycasting(bot);
    }
    
    /**
     * Here we can modify initializing command for our bot.
     *
     * @return
     */
    @Override
    public Initialize getInitializeCommand() {
        Initialize myInit = new Initialize();
        myInit.setLocation(startLocation);
        myInit.setClassName("GameBotsUE2.UDNFemale");
        return myInit;
    }

    /**
     * The bot is initilized in the environment - a physical representation of the
     * bot is present in the game.
     *
     * @param config information about configuration
     * @param init information about configuration
     */
    @Override
    public void botInitialized(GameInfo info, ConfigChange config, InitedMessage init) {
    	// initialize rays for raycasting
        final int rayLength = (int) (UnrealUtils.CHARACTER_COLLISION_RADIUS * 10);
        boolean fastTrace = true;
        boolean floorCorrection = false;
        boolean traceActor = false;

        // 1. remove all previous rays, each bot starts by default with three
        // rays, for educational purposes we will set them manually
        getAct().act(new RemoveRay("All"));

        // 2. create new rays
        raycasting.createRay(LEFT90, new Vector3d(0, -1, 0), rayLength, fastTrace, floorCorrection, traceActor);
        raycasting.createRay(LEFT45, new Vector3d(1, -1, 0), rayLength, fastTrace, floorCorrection, traceActor);
        raycasting.createRay(FRONT, new Vector3d(1, 0, 0), rayLength, fastTrace, floorCorrection, traceActor);
        raycasting.createRay(RIGHT45, new Vector3d(1, 1, 0), rayLength, fastTrace, floorCorrection, traceActor);
        raycasting.createRay(RIGHT90, new Vector3d(0, 1, 0), rayLength, fastTrace, floorCorrection, traceActor);

        // register listener called when all rays are set up in the UT engine
        raycasting.getAllRaysInitialized().addListener(new FlagListener<Boolean>() {

            public void flagChanged(Boolean changedValue) {
                // once all rays were initialized store the AutoTraceRay objects
                // that will come in response in local variables, it is just
                // for convenience
                left = raycasting.getRay(LEFT45);
                front = raycasting.getRay(FRONT);
                right = raycasting.getRay(RIGHT45);
            }
        });

        // 3. declare that we are not going to setup any other rays
        raycasting.endRayInitSequence();
        
        // Most important is that this line ENABLES AUTO TRACE functionality,
        // without ".setAutoTrace(true)" the AddRay command will be useless
        getAct().act(new Configuration().setSpeedMultiplier(moveSpeed).setDrawTraceLines(true).setAutoTrace(true));
    }

    

    /**
     * Main method that controls the bot - makes decisions what to do next.
     * It is called iteratively by Pogamut engine every time a synchronous batch
     * from the environement is received. This is usually 4 times per second - it
     * is affected by visionTime variable, that can be adjusted in GameBots ini file in
     * UT2004/System folder.
     *
     * @throws cz.cuni.amis.pogamut.base.exceptions.PogamutException
     */
    @Override
    public void logic() throws PogamutException {

        if (raycasting.getAllRaysInitialized().getFlag()) {
            // if all rays were initialized then move according to them

            sensorFront = front.isResult();
            sensorLeft = left.isResult();
            sensorRight = right.isResult();

            // is any of the sensor signalig?
            sensor = sensorFront || sensorLeft || sensorRight;

            if (sensor) {
                // if some sensor is signalizing...
                if (moving) {
                    // and we're moving ...
                    loc.stopMovement();
                    moving = false;
                }
                // according to the signals, take action...
                // 8 cases that might happen follows
                if (sensorFront) {
                    if (sensorLeft) {
                        if (sensorRight) {
                            // LEFT, RIGHT, FRONT are signalizing
                            loc.turnHorizontal(bigTurn);
                        } else {
                            // LEFT, FRONT are signalizing
                            loc.turnHorizontal(smallTurn);
                        }
                    } else {
                        if (sensorRight) {
                            // RIGHT, FRONT are signalizing
                            loc.turnHorizontal(-smallTurn);
                        } else {
                            // FRONT is signalizint
                            loc.turnHorizontal(smallTurn);
                        }
                    }
                } else {
                    if (sensorLeft) {
                        if (sensorRight) {
                            // LEFT, RIGHT are signalizing
                            goForward();
                        } else {
                            // LEFT is signalizing
                            loc.turnHorizontal(smallTurn);
                        }
                    } else {
                        if (sensorRight) {
                            // RIGHT is signalizing
                            loc.turnHorizontal(-smallTurn);
                        } else {
                            // no sensor is signalizing
                            goForward();
                        }
                    }
                }
            } else {
                goForward();
            }
        }
    }

    private void goForward() {
        loc.moveContinuos();
        moving = true;
    }

    /**
     * Called each time our bot die. Good for reseting all bot state dependent variables.
     *
     * @param event
     */
    @Override
    public void botKilled(BotKilled event) {
    }

    public static void main(String args[]) throws PogamutException {
    	new SingleUT2004BotRunner<UT2004Bot>(RaycastingBot.class, "RaycastingBot").startAgent();
    }
}
